import { type RPCMessagePortTransportOptions } from "../index.js";
/**
 * Options for the iframe transport.
 */
export type RPCIframeTransportOptions = Omit<RPCMessagePortTransportOptions, "transportId" | "remotePort"> & {
    /**
     * An identifier for the transport. This is used to match the iframe with
     * the parent window. If not set, a default value will be used when establishing
     * the connection.
     */
    transportId?: string | number;
    /**
     * The target origin of the iframe. This is used to restrict the domains
     * that can communicate with the iframe. If not set, the iframe will accept
     * messages from any origin. [Learn more about `targetOrigin` on MDN.](
     * https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage#targetorigin)
     *
     * @default "*"
     */
    targetOrigin?: string;
};
/**
 * Creates a transport to communicate with an iframe. This is an asynchronous
 * process because it requires waiting for the iframe to load and signal that
 * it's ready to receive messages.
 *
 * The target iframe must use `createIframeParentTransport` with a matching
 * `transportId` or `filter` option (if set).
 *
 * Uses `MessageChannel` under the hood. [Learn more about the Channel Messaging
 * API on MDN.](https://developer.mozilla.org/en-US/docs/Web/API/Channel_Messaging_API/Using_channel_messaging)
 *
 * Using the `transportId` option is recommended to
 * avoid conflicts with other connections. If security is a concern, the
 * `targetOrigin` option should be set to the expected origin of the iframe.
 * [Learn more about `targetOrigin` on MDN.](
 * https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage#targetorigin).
 *
 * @example
 *
 * ```ts
 * const rpc = createRPC<ParentSchema, IframeSchema>({
 *   transport: await createIframeTransport(iframeElement, {
 *     transportId: "my-iframe"
 *   }),
 *   // ...
 * });
 * ```
 *
 */
export declare function createIframeTransport(
/**
 * The iframe element to communicate with.
 */
iframe: HTMLIFrameElement, 
/**
 * Options for the iframe transport.
 */
options?: RPCIframeTransportOptions): Promise<import("../types.js").RPCTransport>;
/**
 * Options for the iframe parent transport.
 */
export type RPCIframeParentTransportOptions = Omit<RPCMessagePortTransportOptions, "transportId" | "remotePort"> & {
    /**
     * An identifier for the transport. This is used to match the iframe with
     * the parent window. If not set, a default value will be used when establishing
     * the connection.
     */
    transportId?: string | number;
};
/**
 * Creates a transport to communicate with the parent window from an iframe. This
 * is an asynchronous process because it requires waiting for the parent window
 * to connect to the iframe.
 *
 * The parent window must use `createIframeTransport` with a matching
 * `transportId` or `filter` option (if set).
 *
 * Using the `transportId` option is recommended to avoid conflicts with other
 * connections.
 *
 * @example
 *
 * ```ts
 * const rpc = createRPC<IframeSchema, ParentSchema>({
 *   transport: await createIframeParentTransport({ transportId: "my-iframe" }),
 *   // ...
 * });
 * ```
 */
export declare function createIframeParentTransport(options?: RPCIframeParentTransportOptions): Promise<import("../types.js").RPCTransport>;
